% For use with Matlab
% Copyright (C) 2020 -- Jerome Mendes <jermendes@gmail.com> or <jermendes@isr.uc.pt>, Rui Araujo <rui@isr.uc.pt>
% Jerome Mendes, Ricardo Maia, Rui Araujo, and  Francisco A.A. Souza. "Self-Evolving Fuzzy Controller Composed of Univariate Fuzzy Control". Applied Sicences, 2020.
% The "uSelf-FLC toolbox" comes with ABSOLUTELY NO WARRANTY;
% In case of publication of any application of this method, please, cite the work:
% Jerome Mendes, Ricardo Maia, Rui Araujo, and  Francisco A.A. Souza.
% Self-Evolving Fuzzy Controller Composed of Univariate Fuzzy Control Rules.
% Applied Sciences, 2020, 10(17), 5836.
% DOI: https://doi.org/10.3390/app10175836
% http://www.isr.uc.pt/~jermendes/    ;     http://www.isr.uc.pt/~rui/

function command = uSelfFLC(dinput1, dinput2, error, uk_1, gamma, T_M, delta, Cmin, Cmax)
% inputs                       1        2       3      4      5     6     7    8      9  

%% Description
% This function is the core function to self-evolve the direct fuzzy controller

%% Function inputs
% dinput1 -  values of the input variable x1
% dinput2 -  values of the input variable x2
% error - is the tracking error, i.e. r(k)-y(k)
% uk_1 - previous comamnd signal, u(k-1)
% gamma - is the adaptation gain
% T_M - Size of the temporal sliding window
% delta - Threshold \delta (Criterion 1)
% Cmin - Minimum of the control variable 
% Cmax -Maximum of the control variable 

%% Function output
% command -  fuzzy control signal - equation 7
        
%% Global variables
global safe_interval epsilon Evolve_Parameters x summed_error_window dif_controller delta_dif_controller data_window universe_x;
% safe_interval - minimum number of sample times between the addition of control rules
% epsilon - Thresold of Criteria 3 and 4
% Evolve_Parameters - struct to save some importante parameters to do final plots
% x - Struct with all fuzzy control parameters
% summed_error_window - Sum of estimated control error
% dif_controller - difference of the estimated control error
% delta_dif_controller  - Variation of dif_controller
% data_window - temporal sliding window
% universe_x - Minimum and maximum values of the inputs variables

%% Variables definition
% data_inputs x1 and x2
data_inputs(1,1) = dinput1;
data_inputs(1,2) = dinput2;
% control_min_max - Universe of discourse of the control variable [min max]
control_min_max(1,1) = Cmin;
control_min_max(1,2) = Cmax;

xN_var = size(x, 2); % number of input variables

%% Protection - Values limited to the  Universe of discourse
for j = 1 : xN_var
    if data_inputs(j) <universe_x(j,1)
        data_inputs(j)=universe_x(j,1);
    end
    if data_inputs(j) > universe_x(j,2)
        data_inputs(j) = universe_x(j,2);
    end
end

%% Update sliding window
update_window( data_inputs(2), uk_1);

psi = Psi( data_inputs);

%% Poprosed uSelf-FLC method
command = EvolveFLC();



%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                   FUNCTIONS DEFINITION -- BEGIN                                                       %%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: update_window     %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function update_window( yk, uk)
        %% Description
        % This function updates the temporal sliding window
        
        %% Function inputs
        % yk - output variable of the plant to be controlled
        % uk - command signal
        
        actual_size_window = size(data_window, 2); % size of the temporal sliding window
        
        new_data = [yk; data_inputs(2:end)'; data_inputs(1); uk]; % data to be stored on the temporal sliding window - [ yk+1; variables without ref; ref; uk]
        
        %% Update the temporal sliding window
        if actual_size_window < T_M % if the temporal sliding window is not filled yet
            data_window = [data_window new_data];
        else% if the temporal sliding window is filled
            data_window = [data_window(:,2:end) new_data];
        end
        
    end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: EvolveFLC         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function [command] = EvolveFLC()
        %% Description
        % This function is the core function to self-evolve the fuzzy controller
        
        %% Function output
        % command -  fuzzy control signal - equation 7
        
        %% Variables Definition
        actual_size_window = size(data_window, 2); % Size of the temporal sliding window
        Add = 0;  % Auxiliary variable
        
        %% Function to update the consequent parameters (Algorithm 1 Step 10)
        UpdateConsequent();
        
        %% If the temporal sliding window is filled (Algorithm 1 Step 11)
        if actual_size_window == T_M
            
            %% Function to obtain the estimated control error (equation (13), Algorithm 1 Step 12)
            [error_controller_estimation_all, error_controller_estimation_local] = FuzzyController_EstimationError();
            
            error_controller_estimation = abs(sum(error_controller_estimation_all)); % estimated control error
            
            dif_controller = [ dif_controller abs(error_controller_estimation - summed_error_window )]; % save the difference of the estimated control error
            
            %% If there are enough samples
            if size(dif_controller,2)>2
                
                delta_dif_controller = [delta_dif_controller dif_controller(1,end)-dif_controller(1,end-1)];
                
                %% If Criterion 1 is met (Algorithm 1 Step 13)
                if  delta_dif_controller(1,end) > delta
                    
                    %% Find the input variables which has the maximal estimated control error
                    max_error_var = max(abs(error_controller_estimation_local'));
                    j_E = find(max_error_var == max(max_error_var));
                    
                    Add_aux = 0;
                    for j_c = 1:length(j_E)
                        
                        %% This function obtains the position of the candidate membership function for variable j_E (Equation (16))                        
                        [MF_center, nearestMF] = LocateMF_Gravity(error_controller_estimation_all, j_E(j_c));
                        
                        %% This function, if the Criterion 2 is met, adds a new  membership function on input variable j_E
                        [Add, rule] = AddMF_ErrorAVG_newCons(nearestMF,j_E(j_c), MF_center);
                        
                        Add_aux = Add_aux + Add;
                        
                    end
                    
                    Add = Add_aux;
                    
                end
            end
            
            %% Save the estimated control error
            summed_error_window = abs(sum(error_controller_estimation_all));
        end
        
        %% This function obtains the fuzzy control signal - equation 7
        [command] = FuzzyController ();
        
        %% Save some importante parameters to do plots
        Evolve_Parameters.Add = [Evolve_Parameters.Add Add]; % Flag Add
        Evolve_Parameters.Sample = Evolve_Parameters.Sample +1;
        for j_E=1:size(x,2)% for all input variables
            Evolve_Parameters.nRules{j_E} = [Evolve_Parameters.nRules{j_E} size(x(j_E).Rules,1)];% number of rules
            Evolve_Parameters.MFs{j_E,Evolve_Parameters.Sample} = x(j_E).Rules; % rules matrix (parameters)
        end
        
        %% Delete control rules, if Criteria are met
        delete_condition(); %  Algorithm 1 Step 22)
        
    end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: delete_condition         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function delete_condition()
        %% Description
        % This function delete the fuzzy control rules,if Criteria 3 and 4 are met
        
        %% Variables Definition
        n_var= size(x, 2); % number of inputs variables
        
        %% Auxiliary variable
        k = 1;
        del_aux = 0;
        
        %% Check Criteria 3 and 4
        for j_dc = 1:n_var % for all input variables
            
            n_rules = size( x(j_dc).Rules,1 ); % number of fuzzy rules on input variable j_dc
            del = zeros (1,n_rules); % initializaion (delete if 1)
            
            for i = 1:n_rules % for all fuzzy rules
                
                N_i_ant = x(j_dc).Rules(i,5); % Criterion 3 - the indicator of the total activation degrees (antecedent values) associated to the i_j-th fuzzy rule
                
                x(j_dc).Rules(i,5) = N_i_ant + psi(k); % Criterion 3 - Save the indicator of the total activation degrees (antecedent values) associated to the i_j-th fuzzy rule
                
                if psi(k) == min(psi) % Criterion 4
                    x(j_dc).Rules(i,7) = x(j_dc).Rules(i,7) + 1; % Count if the antecedent value of i_j-th fuzzy is the minimum among all the antecedent values
                end
                
                sample_ini = x(j_dc).Rules(i,6); % k^{ini}_{i_j} - sample time at which the fuzzy rule was created
                
                C1_d =  x(j_dc).Rules(i,5)/(Evolve_Parameters.Sample - sample_ini); % Criterion 3
                C2_d =  1 - x(j_dc).Rules(i,7)/(Evolve_Parameters.Sample - sample_ini);  % Criterion 4
                 
                if C1_d < epsilon
                    
                    if C2_d < epsilon/2
                        
                        if Evolve_Parameters.Sample - sample_ini > safe_interval
                            
                            del(1,i) = 1; % Delete if Ctireria 3 and 4, and the safe_intervalare met
                            
                        end
                        
                    end
                end
                
                k = k + 1;
            end
            
            for i = 1:n_rules
                if del(1,i)==1
                    DeleteRule(j_dc, i); % Delete rule i of variavel j_dc
                    
                    del_aux = del_aux -1;
                end
            end
            
        end
        
        
        
        Evolve_Parameters.del = [Evolve_Parameters.del del_aux]; % Save information about the delete process
    end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: DeleteRule               %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function  [] = DeleteRule(j_Del, i_del)
        %% Description
        % This function deletes the i_del-th fuzzy control rules of input
        % variable j_Del
        
        %% Function inputs
        % j_Del - input variable in which a control rule willbe deleted
        % i_del - index of the control rule to be deleted
                
        x(j_Del).Rules(i_del-1,3) = x(j_Del).Rules(i_del+1,2); % Update the previous MF       
        x(j_Del).Rules(i_del+1,1) = x(j_Del).Rules(i_del-1,2); % Update the next MF
        
        x(j_Del).Rules(i_del,:) = []; % Delete control rule
        
    end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: Psi                      %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function [psi_aux] = Psi(data)
        %% Description
        % This function obtains the variable \Psi (equation (4) )
        
        %% Function inputs
        % data -  values of the input variable
        
        %% Function output
        % psi_aux - variable \psi - equation (4)
        
        n_var= size(x, 2); % number of inputs variables
        k = 1;
        
        for j_psi=1:n_var % for all input variables
            
            n_rules = size( x(j_psi).Rules,1 ); % number of fuzzy rules on input variable j_psi
            
            for i = 1:n_rules % for all fuzzy rules
                
                % MF parameters of the fuzzy rule i of input variable j_psi
                parameters(1,1) = x(j_psi).Rules(i,1); % lower limit
                parameters(1,2) = x(j_psi).Rules(i,2); % center
                parameters(1,3) = x(j_psi).Rules(i,3); % upper limit
                
                mu_varialble(k) = trifp(data(j_psi), parameters); % fuzzy degree on the MF of the fuzzy rule i of input variable j_psi
                k = k + 1;
            end
        end
        
        psi_aux = mu_varialble / (sum(mu_varialble) + 0.000000001); % \psi - equation 4 (0.000000001 is just a protection....)
        
    end


%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: trifp                    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function u_mf = trifp(data_mf, parameters_trip)
        %% Description
        % This function obtains the fuzzy degree of a triangular MF for a given data
        
        %% Function inputs
        % data_mf - data
        % parameters - parameters of the triangular MF - parameters = [a b c] which
        % are,  the lower limit, center value, and upper limit of the MF,
        % respectively.
        
        %% Function output
        % u_mf - fuzzy degree of a triangular MF
        
        u_mf = 0;
        
        if (data_mf> parameters_trip(1) && data_mf< parameters_trip(2)) % left size
            u_mf = (data_mf-parameters_trip(1))/(parameters_trip(2)-parameters_trip(1));
        elseif (data_mf>= parameters_trip(2) && data_mf< parameters_trip(3))% right size
            u_mf = (parameters_trip(3)-data_mf)/(parameters_trip(3)-parameters_trip(2));
        elseif (data_mf<= parameters_trip(1) || data_mf>= parameters_trip(3)) % outside
            u_mf = 0;
        end
        
    end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: UpdateConsequent         %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function UpdateConsequent()
        %% Description
        % This function updates the consequent parameters (Algorithm 1 Step 10)
           
        %% Variables Definition
        n_var= size(x, 2); % number of inputs variables
        
        
        %% Consequents update
        k = 1; % Auxiliary variable
        
        for j_UC=1:n_var % for all input variables
            
            n_rules = size( x(j_UC).Rules,1 ); % number of rules of input variable j_UC
            
            for i = 1:n_rules % for all rules of input variable j_UC,
                
                x(j_UC).Rules(i,4) = x(j_UC).Rules(i,4) + gamma*psi(k)*error;  % adapt the consequent of rule i
                
                if x(j_UC).Rules(i,4) < control_min_max(1)
                    x(j_UC).Rules(i,4)  = control_min_max(1);  % Limitation of the consequent of rule i by the minimal control value admissible
                end
                
                if x(j_UC).Rules(i,4) > control_min_max(2)
                    x(j_UC).Rules(i,4) = control_min_max(2); % Limitation of the consequent of rule i by the maximal control value admissible
                    
                end
                
                k = k +1;
            end
            
        end
        
    end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: FuzzyController_EstimationError   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function [error_controller_estimation, error_controller_estimation_local] = FuzzyController_EstimationError ()
        %% Description
        % This function obtains the estimated control error (equation (13), see Section 3.3.2)
        
        %% Function output
        % error_controller_estimation - estimated control error, equation (13)
        % error_controller_estimation_local - estimated control error for each rule
        
        %% Variables Definition
        n_var= size(x, 2); % number of inputs variables
        error_controller_estimation_local = zeros(n_var,T_M); % Memory allocation with zeros
        error_controller_estimation = zeros(1,T_M);% Memory allocation with zeros
        
        for m=1:T_M% for all temporal sliding window
            
            command_signal = 0.0; % initialization
            
            %% Obtain variable \psi for temporal sliding window
            psi_window = Psi( data_window(1:end-2,m));
            
            k = 1;
            
            for j_FE=1:n_var % for all input variables
                
                n_rules = size( x(j_FE).Rules,1 );% number of rules of input variable j_FE
                
                local_error_variable = 0;% initialization
                
                for i = 1:n_rules % for all rules of input variable j_FE,
                    
                    command_signal = command_signal + psi_window(k)*x(j_FE).Rules(i,4); % Obtain the estimated controller signal using the current fuzzy controller
                    
                    local_error_variable = local_error_variable + abs(data_window(end,m)*psi_window(k) - psi_window(k)*x(j_FE).Rules(i,4)); % Estimation control error of rule i of variable j_FE
                    
                    k = k +1;
                end
                
                error_controller_estimation_local(j_FE,m) = local_error_variable; % Save the estimation control error of each input variable
                
            end
            
            error_controller_estimation(m) = data_window(end,m) - command_signal; % estimated control error for all window
            
        end
        
    end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: LocateMF_Gravity                  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function [MF_center, nearestMF] = LocateMF_Gravity(error_controller_estimation_all, j_LG)
        %% Description
        % This function obtains the position/center of the new (candidate) membership function for variable j (Equation (16))
        
        %% Function inputs
        % error_controller_estimation_all - estimated control error for each rule
        % j_LG - input variable which has the maximal estimated control error (i.e. where the new MF will be added)
        
        %% Function output
        % MF_center - position/center of the new (candidate) membership function
        % nearestMF -  the nearest membership function from candidate membership function
        
        error_size = size(error_controller_estimation_all, 2);% Memory allocation with zeros
        
        %%  Distribution of control estimation error - equation (22)
        
        num = 0.0;  % Auxiliary variable
        den = 0.0;  % Auxiliary variable
        
        for m=1:error_size % for all temporal sliding window
            
            num = num + m*error_controller_estimation_all(m)^2; % numerator of equation (16)
            den = den + error_controller_estimation_all(m)^2; % denominator of equation (16)
        end
        
        gravity_center_error = num/den; % equation (16)
        
        if j_LG == 1
            var = size(data_window, 1)-1; % if the selected variable is ref (j=1), this variable is stored at end-1 at data_window
        else
            var = j_LG;
        end
        
        MF_center =  data_window(var, ceil(gravity_center_error) ); % position/center of the new (candidate) membership function
        
        %% Obtain the distance between the center of the candidate membership function and the other ones
        n_rules = size( x(j_LG).Rules,1 );% number of rules on input variable j
        dist = zeros(1,n_rules);% Memory allocation with zeros
        
        for i=1:n_rules
            dist(i) = abs(x(j_LG).Rules(i,2) - MF_center);
        end
        
        nearestMF = find( dist == min(dist), 1 ); % the nearest membership function from candidate membership function
        
    end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: AddMF_ErrorAVG           %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function  [Add, rule] = AddMF_ErrorAVG_newCons(nearestMF, j_Add, MF_center)
        %% Description
        % In this function, if the Criterion 2 is met, a new  membership
        % function is added on input variable j_Add
        
        %% Function inputs
        % nearestMF - the nearest membership function from candidate membership function
        % j_Add - input variable which has the maximal estimated control error (i.e. where the new MF will be added)
        % MF_center - position/center of the new (candidate) membership function
        
        %% Function output
        % Add - Flag to indicate if the candidate membership function was added
        % (Add=1) or not (Add=0)
        % rule - index of the new rule
        
        %% Auxiliary variables
        n_rules = size( x(j_Add).Rules,1 ); % number of rules on input variable j_Add
        tau = x(j_Add).eta; % the minimal distance beetween MFs for input variable j_Add
        rule = -1;
        Add = 0;
        
        %MF_center
        %% Case 1 - If the candidate membership function is not the first  or last membership function
        if nearestMF ~= 1 && nearestMF < n_rules
            
            b_MostActive = x(j_Add).Rules( nearestMF, 2 ); % MF's center of nearest MF  from candidate MF
            b_previous = x(j_Add).Rules( nearestMF-1, 2 );% MF's center of the previous MF of nearest MF
            b_next = x(j_Add).Rules( nearestMF+1, 2 );% MF's center of the next MF of nearest MF
            
            dist = abs(MF_center - b_MostActive) / 2; % distance beetween the candidate and nearest MFs
            dist_previous = abs(MF_center - b_previous) / 2; % distance beetween the candidate and the previous MF of nearest MFs
            dist_next = abs(MF_center - b_next) / 2;  % distance beetween the candidate and the next MF of nearest MFs
            
            %% Criterion 2
            if dist > tau % If the distance beetween the candidate and nearest MFs is bigger than the  minimal distance admissible
                
                if MF_center < b_MostActive && dist_previous > tau
                    
                    Add = 1; % The candidate membership function will be added - Criterion 2 is met
                    
                    % new consequent %
                    left_MF = nearestMF - 1;
                    right_MF = nearestMF;
                    parameters(1,1) = x(j_Add).Rules(left_MF,1);
                    parameters(1,2) = x(j_Add).Rules(left_MF,2);
                    parameters(1,3) = x(j_Add).Rules(left_MF,3);
                    conseq_l = x(j_Add).Rules(left_MF,4);
                    
                    mu_left = trifp(MF_center, parameters);
                    
                    parameters(1,1) = x(j_Add).Rules(right_MF,1);
                    parameters(1,2) = x(j_Add).Rules(right_MF,2);
                    parameters(1,3) = x(j_Add).Rules(right_MF,3);
                    conseq_r = x(j_Add).Rules(right_MF,4);
                    
                    mu_right = trifp(MF_center, parameters);
                    
                    new_b =  MF_center; % center of the MF
                    new_left =  x(j_Add).Rules( left_MF, 2 ); % lower limit of the new MF
                    new_right =  x(j_Add).Rules( right_MF, 2 ); % upper limit of the new MF
                    
                    % Changes on the left MF of the new MF
                    x(j_Add).Rules( left_MF, 3 ) = new_b;
                    
                    % Changes on the right MF of the new MF
                    x(j_Add).Rules( right_MF , 1 ) = new_b;
                    
                    % update/sort the Rule Struct
                    for i = n_rules:-1:nearestMF
                        [x(j_Add).Rules( i+1,:)] = x(j_Add).Rules( i,: );
                    end
                    
                    
                    % Insert the new rule
                    x(j_Add).Rules(nearestMF, 1 ) = new_left;
                    x(j_Add).Rules(nearestMF, 2 ) = new_b;
                    x(j_Add).Rules(nearestMF, 3 ) = new_right;
                    x(j_Add).Rules(nearestMF, 4 ) =  mu_left * conseq_l+ mu_right * conseq_r;
                    x(j_Add).Rules(nearestMF, 5 ) = 0;
                    x(j_Add).Rules(nearestMF, 6 ) = Evolve_Parameters.Sample;
                    x(j_Add).Rules(nearestMF, 7 ) = 0;
                    
                    rule = nearestMF; % index of the new rule
                end
                if MF_center > b_MostActive && dist_next > tau
                    
                    Add = 1;% The candidate membership function will be added - Criterion 2 is met
                    
                    % new consequent %
                    left_MF = nearestMF;
                    right_MF = nearestMF+1;
                    parameters(1,1) = x(j_Add).Rules(left_MF,1);
                    parameters(1,2) = x(j_Add).Rules(left_MF,2);
                    parameters(1,3) = x(j_Add).Rules(left_MF,3);
                    conseq_l = x(j_Add).Rules(left_MF,4);
                    
                    mu_left = trifp(MF_center, parameters);
                    
                    parameters(1,1) = x(j_Add).Rules(right_MF,1);
                    parameters(1,2) = x(j_Add).Rules(right_MF,2);
                    parameters(1,3) = x(j_Add).Rules(right_MF,3);
                    conseq_r = x(j_Add).Rules(right_MF,4);
                    
                    mu_right = trifp(MF_center, parameters);
                    
                    new_b =  MF_center;% center of the MF
                    new_left = x(j_Add).Rules( nearestMF , 2 );% lower limit of the new MF
                    new_right =  x(j_Add).Rules( nearestMF + 1 , 2 );% upper limit of the new MF
                    
                    % Changes on the previous MF of the new MF
                    x(j_Add).Rules( nearestMF , 3 ) = new_b;
                    
                    % Changes on next MF of the new MF
                    x(j_Add).Rules( nearestMF +1, 1 ) = new_b;
                    
                    % update/sort the Rule Struct
                    for i = n_rules:-1:nearestMF+1
                        [x(j_Add).Rules(i+1,:)] = x(j_Add).Rules(i,:);
                    end
                    
                    % Insert the new rule
                    x(j_Add).Rules(nearestMF+1, 1 ) = new_left;
                    x(j_Add).Rules(nearestMF+1, 2 ) = new_b;
                    x(j_Add).Rules(nearestMF+1, 3 ) = new_right;
                    x(j_Add).Rules(nearestMF+1, 4 ) = mu_left * conseq_l+ mu_right * conseq_r;
                    x(j_Add).Rules(nearestMF+1, 5 ) = 0;
                    x(j_Add).Rules(nearestMF+1, 6 ) = Evolve_Parameters.Sample;
                    x(j_Add).Rules(nearestMF+1, 7 ) = 0;
                    
                    rule = nearestMF+1;% index of the new rule
                end
                
            end
            %% Case 2 - If the candidate membership function is the first membership function
        elseif nearestMF == 1
            
            b_MostActive = x(j_Add).Rules( nearestMF , 2 );% MF's center of nearest MF  from candidate MF
            b_next = x(j_Add).Rules( nearestMF+1 , 2 );% MF's center of the next MF of nearest MF
            
            dist_1 = abs(MF_center - b_MostActive) / 2;% distance beetween the candidate and nearest MFs
            dist_2 = abs(MF_center - b_next) / 2;% distance beetween the candidate and the next MF of nearest MFs
            
            if dist_1 > tau && dist_2 > tau && MF_center > b_MostActive % If the distance beetween the candidate and nearest MFs is bigger than the  minimal distance admissible
                
                Add = 1;% The candidate membership function will be added - Criterion 2 is met
                
                % new consequent %
                left_MF = nearestMF;
                right_MF = nearestMF+1;
                parameters(1,1) = x(j_Add).Rules(left_MF,1);
                parameters(1,2) = x(j_Add).Rules(left_MF,2);
                parameters(1,3) = x(j_Add).Rules(left_MF,3);
                conseq_l = x(j_Add).Rules(left_MF,4);
                
                mu_left = trifp(MF_center, parameters);
                
                parameters(1,1) = x(j_Add).Rules(right_MF,1);
                parameters(1,2) = x(j_Add).Rules(right_MF,2);
                parameters(1,3) = x(j_Add).Rules(right_MF,3);
                conseq_r = x(j_Add).Rules(right_MF,4);
                
                mu_right = trifp(MF_center, parameters);
                
                
                new_b =  MF_center;% center of the MF
                new_left = x(j_Add).Rules( nearestMF , 2 );% lower limit of the new MF
                new_right = x(j_Add).Rules( nearestMF + 1, 2 ); % upper limit of the new MF
                
                % Changes  on the left MF of the new MF
                x(j_Add).Rules( nearestMF , 3 ) = new_b;
                
                % Changes on the right MF of the new MF
                x(j_Add).Rules( nearestMF + 1, 1 ) = new_b;
                
                % update/sort the Rule Struct
                for i = n_rules:-1:nearestMF+1
                    [x(j_Add).Rules(i+1,:)] = x(j_Add).Rules(i,:);
                end
                
                % Insert the new rule
                x(j_Add).Rules(nearestMF+1, 1 ) = new_left;
                x(j_Add).Rules(nearestMF+1, 2 ) = new_b;
                x(j_Add).Rules(nearestMF+1, 3 ) = new_right;
                x(j_Add).Rules(nearestMF+1, 4 ) =  mu_left * conseq_l+ mu_right * conseq_r;
                x(j_Add).Rules(nearestMF+1, 5 ) = 0;
                x(j_Add).Rules(nearestMF+1, 6 ) = Evolve_Parameters.Sample;
                x(j_Add).Rules(nearestMF+1, 7 ) = 0;
                
                rule = nearestMF+1;% index of the new rule
            end
            %% Case 3 - If the candidate membership function is the last membership function
        else
            
            b_MostActive = x(j_Add).Rules( nearestMF , 2 );% MF's center of nearest MF  from candidate MF
            b_previous= x(j_Add).Rules( nearestMF-1, 2 );% MF's center of the previous MF of nearest MF
            
            dist_1 = abs(MF_center - b_MostActive) / 2;% distance beetween the candidate and nearest MFs
            dist_2 = abs(MF_center - b_previous) / 2;% distance beetween the candidate and the previous MF of nearest MFs
            
            if dist_1 > tau && dist_2 > tau   && MF_center < b_MostActive % If the distance beetween the candidate and nearest MFs is bigger than the  minimal distance admissible
                
                Add = 1;% The candidate membership function will be added - Criterion 2 is met
                
                % new consequent %
                left_MF = nearestMF - 1;
                right_MF = nearestMF;
                parameters(1,1) = x(j_Add).Rules(left_MF,1);
                parameters(1,2) = x(j_Add).Rules(left_MF,2);
                parameters(1,3) = x(j_Add).Rules(left_MF,3);
                conseq_l = x(j_Add).Rules(left_MF,4);
                
                mu_left = trifp(MF_center, parameters);
                
                parameters(1,1) = x(j_Add).Rules(right_MF,1);
                parameters(1,2) = x(j_Add).Rules(right_MF,2);
                parameters(1,3) = x(j_Add).Rules(right_MF,3);
                conseq_r = x(j_Add).Rules(right_MF,4);
                
                mu_right = trifp(MF_center, parameters);
                
                
                new_b =  MF_center;% center of the MF
                
                % Changes  on the left MF of the new MF
                x(j_Add).Rules( nearestMF - 1 , 3 ) = new_b;
                
                % Creat the new MF
                new_left = x(j_Add).Rules( nearestMF - 1 , 2 );% lower limit of the new MF
                new_right = x(j_Add).Rules( nearestMF , 2 );% upper limit of the new MF
                
                % Changes  on the right MF of the new MF
                x(j_Add).Rules( nearestMF , 1 ) = new_b;
                
                % update/sort the Rule Struct
                for i = n_rules:-1:nearestMF
                    [x(j_Add).Rules(i+1,:)] = x(j_Add).Rules(i,:);
                end
                
                % Insert the new rule
                x(j_Add).Rules(nearestMF, 1 ) = new_left;
                x(j_Add).Rules(nearestMF, 2 ) = new_b;
                x(j_Add).Rules(nearestMF, 3 ) = new_right;
                x(j_Add).Rules(nearestMF, 4 ) =  mu_left * conseq_l+ mu_right * conseq_r;
                x(j_Add).Rules(nearestMF, 5 ) = 0;
                x(j_Add).Rules(nearestMF, 6 ) = Evolve_Parameters.Sample;
                x(j_Add).Rules(nearestMF, 7 ) = 0;
                
                rule = nearestMF;% index of the new rule
            end
            
        end
        
        %Add
    end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   FUNCTION: FuzzyController            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    function [command_signal] = FuzzyController ()
        %% Description
        % This function obtains the fuzzy control signal
        
        %% Function output
        % command_signal -  fuzzy control signal
        
        %% Variables Definition
        n_var= size(x, 2); % number of inputs variables
        command_signal = 0.0;
        
        %% Obtain variable \psi
        psi = Psi(data_inputs);
        
        k = 1;% Auxiliary variable
        
        for j_FC=1:n_var %  for all input variables
            
            n_rules = size( x(j_FC).Rules,1 ); % number of rules of input variable j
            
            for i = 1:n_rules % for all rules of input variable j
                
                command_signal = command_signal + psi(k)*x(j_FC).Rules(i,4); % fuzzy control signal
                k = k + 1;
            end
            
        end
        
    end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                   FUNCTIONS DEFINITION -- END                                                         %%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

end % end of uSelfFLC
